//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Fit/ParameterTreeItems.cpp
//! @brief     Implements classes for ParameterTreeItems
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Fit/ParameterTreeItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Job/JobModelFunctions.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Material/MaterialModel.h"
#include "GUI/Model/Session/ModelPath.h"
#include "GUI/Model/Session/SessionXML.h"
#include "GUI/Model/Types/VectorDescriptor.h"
#include "GUI/Util/DeserializationException.h"
#include <QXmlStreamReader>
#include <QXmlStreamWriter>

using namespace GUI::Session::XML;

namespace {
namespace Tags {
const QString BackupValues("BackupValues");
const QString BackupValue("BackupValue");
const QString Link("Link");
const QString Value("Value");
} // namespace Tags
} // namespace

// ----------------------------------------------------------------------------

ParameterLabelItem::ParameterLabelItem() : SessionItem(M_TYPE)
{
    static const QString T_CHILDREN = "children tag";
    registerTag(T_CHILDREN, 0, -1, {M_TYPE, ParameterItem::M_TYPE});
    setDefaultTag(T_CHILDREN);
}

// ----------------------------------------------------------------------------

ParameterItem::ParameterItem() : SessionItem(M_TYPE) {}

//! Sets current value to the original PropertyItem of MultiLayerItem/InstrumentItem.

void ParameterItem::propagateValueToLink(double newValue)
{
    setValue(newValue);

    m_d.set(newValue);
}

void ParameterItem::linkToMaterial(MaterialItem* material, DoubleDescriptor d)
{
    m_link = d.path();
    m_d = d;
}

void ParameterItem::linkToSessionItem(SessionItem* item)
{
    m_link = GUI::Model::Path::getPathFromIndex(item->index());
    m_d = DoubleDescriptor(item, "");
}

RealLimits ParameterItem::limitsOfLink() const
{
    return m_d.limits;
}

int ParameterItem::decimalsOfLink() const
{
    return m_d.decimals;
}

QString ParameterItem::link() const
{
    return m_link;
}

// ----------------------------------------------------------------------------

ParameterContainerItem::ParameterContainerItem() : SessionItem(M_TYPE)
{
    static const QString T_CHILDREN = "children tag";
    registerTag(T_CHILDREN, 0, -1, {ParameterLabelItem::M_TYPE});
    setDefaultTag(T_CHILDREN);
}

void ParameterContainerItem::writeNonSessionItems(QXmlStreamWriter* writer) const
{
    writeAttribute(writer, GUI::Session::XML::Version, int(1));

    writer->writeStartElement(Tags::BackupValues);

    for (auto v = m_backupValues.cbegin(); v != m_backupValues.cend(); v++) {
        writer->writeEmptyElement(Tags::BackupValue);
        writeAttribute(writer, Tags::Link, v.key());
        writeAttribute(writer, Tags::Value, v.value());
    }
    writer->writeEndElement();
}

void ParameterContainerItem::readNonSessionItems(QXmlStreamReader* reader)
{
    m_backupValues.clear();
    const int version = reader->attributes().value(GUI::Session::XML::Version).toInt();

    if (version < 1)
        throw DeserializationException::tooOld();

    if (version > 1)
        throw DeserializationException::tooNew();

    while (reader->readNextStartElement()) {
        if (reader->name().toString() == Tags::BackupValues) {

            while (reader->readNextStartElement()) {
                ASSERT(reader->name().toString() == Tags::BackupValue);
                QString link;
                double d;
                readAttribute(reader, Tags::Link, &link);
                readAttribute(reader, Tags::Value, &d);
                m_backupValues[link] = d;
                reader->skipCurrentElement();
            }
        }
    }
}

bool ParameterContainerItem::allowWritingChildToXml(SessionItem* /*child*/) const
{
    // no children are written. The contained tree is always generated
    return false;
}

void ParameterContainerItem::setBackupValue(const QString& link, double d)
{
    m_backupValues[link] = d;
}

void ParameterContainerItem::restoreBackupValue(SessionItem* item)
{
    ASSERT(item);
    if (auto* parameter = dynamic_cast<ParameterItem*>(item))
        if (m_backupValues.contains(parameter->link()))
            parameter->propagateValueToLink(m_backupValues[parameter->link()]);

    for (auto* child : item->children())
        restoreBackupValue(child);
}

void ParameterContainerItem::restoreBackupValues()
{
    for (auto* child : children())
        restoreBackupValue(child);
}
