//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Session/SessionXML.cpp
//! @brief     Implements reader and writer classes for SessionModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Group/GroupItem.h"
#include "GUI/Model/Material/MaterialItem.h"
#include "GUI/Model/Session/SessionItemTags.h"
#include "GUI/Model/Session/SessionModel.h"
#include "GUI/Util/ComboProperty.h"
#include "GUI/Util/DeserializationException.h"
#include "GUI/Util/Error.h"
#include "GUI/Util/MessageService.h"
#include <QColor>
#include <QtCore/QXmlStreamWriter>

namespace {

void report_error(MessageService* messageService, SessionItem* item, const QString& message)
{
    if (!messageService)
        throw Error(QString("Warning: ") + message);
    if (!item)
        throw Error(QString("report_error called without parent; warning: ") + message);
    messageService->addWarning(item->model(), message);
}

SessionItem* createItem(SessionItem* parent, const QString& modelType, const QString& tag)
{
    ASSERT(parent);
    if (parent->hasModelType<GroupItem>()) {
        if (const auto* groupItem = dynamic_cast<const GroupItem*>(parent))
            return groupItem->getItemOfType(modelType);
        return nullptr;
    }
    if (parent->sessionItemTags()->isSingleItemTag(tag))
        return parent->getItem(tag);
    try {
        return parent->model()->insertNewItem(modelType, parent, -1, tag);
    } catch (const std::exception&) {
        return nullptr; // error will be reported later
    }
}

} // namespace


void GUI::Session::XML::writeModel(QXmlStreamWriter* writer, SessionItem* modelRootItem)
{
    ASSERT(writer);
    ASSERT(modelRootItem);
    writer->writeStartElement(modelRootItem->model()->getModelTag());

    writeItemAndChildItems(writer, modelRootItem);

    writer->writeEndElement(); // m_model_tag
}

void GUI::Session::XML::writeItemAndChildItems(QXmlStreamWriter* writer, const SessionItem* item)
{
    ASSERT(item);
    if (item->parent()) {
        writer->writeStartElement(GUI::Session::XML::ItemTag);
        writer->writeAttribute(GUI::Session::XML::ModelTypeAttribute, item->modelType());
        QString tag = item->parent()->tagFromItem(item);
        writer->writeAttribute(GUI::Session::XML::TagAttribute, tag);
        writer->writeAttribute(GUI::Session::XML::DisplayNameAttribute,
                               item->roleProperty(SessionFlags::DisplayNameRole).toString());
        for (int role : item->getRoles()) {
            if (role == Qt::DisplayRole || role == Qt::EditRole)
                GUI::Session::XML::writeVariant(writer, item->value(), role);
        }
    }

    for (auto* child : item->children())
        if (item->allowWritingChildToXml(child))
            writeItemAndChildItems(writer, child);

    QByteArray a = item->serializeBinaryData();
    if (!a.isEmpty()) {
        writer->writeStartElement(GUI::Session::XML::BinaryData);
        writer->writeAttribute(GUI::Session::XML::Version, "1");
        writer->writeCharacters(a.toBase64());
        writer->writeEndElement();
    }

    QString nonSessionItemContent;
    QXmlStreamWriter nonSessionItemStream(&nonSessionItemContent);
    nonSessionItemStream.writeStartElement(
        GUI::Session::XML::NonSessionItemData); // for following writeAttrib
    const auto sizeBefore = nonSessionItemContent.size();
    item->writeNonSessionItems(&nonSessionItemStream);
    if (sizeBefore != nonSessionItemContent.size()) {
        writer->writeStartElement(GUI::Session::XML::NonSessionItemData);
        item->writeNonSessionItems(writer);
        writer->writeEndElement();
    }

    if (item->parent())
        writer->writeEndElement(); // ItemTag
}

void GUI::Session::XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeName,
                                       const QVariant& variant)
{
    ASSERT(writer);
    ASSERT(variant.isValid());

    if (variant.type() == QVariant::Double) {
        writeAttribute(writer, attributeName, variant.toDouble());
    } else if (variant.type() == QVariant::Int) {
        writeAttribute(writer, attributeName, variant.toInt());
    } else if (variant.type() == QVariant::UInt) {
        writeAttribute(writer, attributeName, variant.toUInt());
    } else if (variant.type() == QVariant::Bool) {
        writer->writeAttribute(attributeName, QString::number(variant.toBool()));
    } else if (variant.type() == QVariant::String) {
        writer->writeAttribute(attributeName, variant.toString());
    } else if (variant.type() == QVariant::Color) {
        const auto col = variant.value<QColor>();
        const QString tcol = col.isValid() ? col.name(QColor::HexArgb) : "";
        writer->writeAttribute(attributeName, tcol);
    } else if (QString(variant.typeName()) == "ComboProperty") {
        auto combo = variant.value<ComboProperty>();
        writer->writeAttribute(attributeName, combo.stringOfSelections());
        writer->writeAttribute(GUI::Session::XML::ParameterExtAttribute, combo.stringOfValues());

    } else
        throw Error("GUI::Session::XML::writeVariant: Parameter type not supported "
                    + QString(variant.typeName()));
}

void GUI::Session::XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeBaseName,
                                       const R3& vec)
{
    writeAttribute(writer, attributeBaseName + "X", vec.x());
    writeAttribute(writer, attributeBaseName + "Y", vec.y());
    writeAttribute(writer, attributeBaseName + "Z", vec.z());
}

void GUI::Session::XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeBaseName,
                                       const complex_t& c)
{
    writeAttribute(writer, attributeBaseName + "Re", c.real());
    writeAttribute(writer, attributeBaseName + "Im", c.imag());
}

void GUI::Session::XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeName,
                                       double d)
{
    writer->writeAttribute(attributeName, d == 0.0 ? "0" : QString::number(d, 'e', 12));
}

void GUI::Session::XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeName,
                                       int d)
{
    writer->writeAttribute(attributeName, QString::number(d));
}

void GUI::Session::XML::writeAttribute(QXmlStreamWriter* writer, const QString& attributeName,
                                       unsigned d)
{
    writer->writeAttribute(attributeName, QString::number(d));
}

void GUI::Session::XML::writeVariant(QXmlStreamWriter* writer, QVariant variant, int role)
{
    ASSERT(writer);
    if (variant.isValid()) {
        writer->writeStartElement(GUI::Session::XML::ParameterTag);
        writer->writeAttribute(GUI::Session::XML::ParameterTypeAttribute, variant.typeName());
        writer->writeAttribute(GUI::Session::XML::ParameterRoleAttribute, QString::number(role));
        writeAttribute(writer, GUI::Session::XML::ParameterValueAttribute, variant);
        writer->writeEndElement(); // end ParameterTag
    }
}

void GUI::Session::XML::readItems(QXmlStreamReader* reader, SessionItem* parent, QString topTag,
                                  MessageService* messageService)
{
    bool legacyDistributionFound = false;
    ASSERT(reader);
    ASSERT(parent);
    const QString start_type = parent->model()->getModelTag();
    while (!reader->atEnd()) {
        reader->readNext();
        if (reader->isStartElement()) {
            if (reader->name() == GUI::Session::XML::ItemTag) {
                const QString model_type =
                    reader->attributes().value(GUI::Session::XML::ModelTypeAttribute).toString();

                if (model_type == "ParticleDistribution") {
                    if (!legacyDistributionFound) {
                        QString message(
                            "This file contains particle distributions in the sample. Elements of "
                            "this type are not supported any more. Reading the file will be "
                            "continued as much as possible, but you should review the sample.");

                        report_error(messageService, parent, message);
                    }
                    legacyDistributionFound = true;
                    reader->skipCurrentElement();
                    continue;
                }

                QString tag =
                    reader->attributes().value(GUI::Session::XML::TagAttribute).toString();
                QString displayName =
                    reader->attributes().value(GUI::Session::XML::DisplayNameAttribute).toString();

                if (!topTag.isEmpty()) {
                    // to handle copying of item to another parent
                    tag = topTag;
                    topTag.clear();
                }
                auto* newItem = createItem(parent, model_type, tag);
                if (!newItem) {
                    QString message = QString("Error while parsing XML. Can't create item of "
                                              "modelType '%1' for tag '%2', parent '%3'")
                                          .arg(model_type, tag, parent->itemName());
                    report_error(messageService, parent, message);
                    // risky attempt to recover the rest of the project
                    reader->skipCurrentElement();
                } else {
                    parent = newItem;
                    parent->setDisplayName(displayName);
                }
                ASSERT(reader->name() == GUI::Session::XML::ItemTag);
            } else if (reader->name() == GUI::Session::XML::ParameterTag) {
                GUI::Session::XML::readProperty(reader, parent, messageService);
                ASSERT(reader->name() == GUI::Session::XML::ParameterTag);
            } else if (reader->name() == GUI::Session::XML::BinaryData) {
                if (reader->attributes().value(GUI::Session::XML::Version).toInt() == 1) {
                    QString valueAsBase64 =
                        reader->readElementText(QXmlStreamReader::SkipChildElements);
                    const auto data = QByteArray::fromBase64(
                        valueAsBase64.toLatin1()); // #baimport add a unit test for this!
                    parent->deserializeBinaryData(data);
                } else
                    throw DeserializationException::tooNew();
                ASSERT(reader->name() == GUI::Session::XML::BinaryData);
            } else if (reader->name() == GUI::Session::XML::NonSessionItemData) {
                parent->readNonSessionItems(reader);
                ASSERT(reader->name() == GUI::Session::XML::NonSessionItemData);
            }
        } else if (reader->isEndElement()) {
            if (reader->name() == GUI::Session::XML::ItemTag && parent)
                parent = parent->parent();
            if (reader->name() == start_type)
                break;
        }
    }
}

QString GUI::Session::XML::readProperty(QXmlStreamReader* reader, SessionItem* item,
                                        MessageService* messageService)
{
    ASSERT(reader);
    const QString parameter_name =
        reader->attributes().value(GUI::Session::XML::ParameterNameAttribute).toString();
    const QString parameter_type =
        reader->attributes().value(GUI::Session::XML::ParameterTypeAttribute).toString();
    const int role = reader->attributes().value(GUI::Session::XML::ParameterRoleAttribute).toInt();

    if (!item) {
        QString message =
            QString("Attempt to set property '%1' for non existing item").arg(parameter_name);
        report_error(messageService, item, message);
        return parameter_name;
    }

    QVariant variant;
    if (parameter_type == "double") {
        double parameter_value =
            reader->attributes().value(GUI::Session::XML::ParameterValueAttribute).toDouble();
        variant = parameter_value;
    } else if (parameter_type == "int") {
        int parameter_value =
            reader->attributes().value(GUI::Session::XML::ParameterValueAttribute).toInt();
        variant = parameter_value;
    } else if (parameter_type == "uint") {
        unsigned parameter_value =
            reader->attributes().value(GUI::Session::XML::ParameterValueAttribute).toUInt();
        variant = parameter_value;
    } else if (parameter_type == "bool") {
        bool parameter_value =
            reader->attributes().value(GUI::Session::XML::ParameterValueAttribute).toInt();
        variant = parameter_value;
    } else if (parameter_type == "QString") {
        QString parameter_value =
            reader->attributes().value(GUI::Session::XML::ParameterValueAttribute).toString();
        variant = parameter_value;
    } else if (parameter_type == "QColor") {
        QString parameter_value =
            reader->attributes().value(GUI::Session::XML::ParameterValueAttribute).toString();
        variant = QColor(parameter_value);
    } else if (parameter_type == "ExternalProperty") {

        const bool isMaterialItemColor =
            item->parent() != nullptr && dynamic_cast<MaterialItem*>(item->parent());

        if (isMaterialItemColor) {
            // This handles legacy XML files for MaterialItem:
            // In former times, the color of a MaterialItem was held as an "ExternalProperty". This
            // has been changed to a simple QColor property => use only the color value of the
            // "ExternalProperty" and store it as QColor in the variant
            const QString colorName =
                reader->attributes().value(GUI::Session::XML::ExternalPropertyColorAtt).toString();
            variant = QColor(colorName);
        } else {
            // This handles legacy XML files for items referring to a material:
            // In former times, the material in e.g. a ParticleItem was held as an
            // "ExternalProperty". This has been changed to the material's identifier only (as
            // QString)
            // => use only the identifier value of the "ExternalProperty" and store it as QString in
            // the variant
            const QString identifier = reader->attributes()
                                           .value(GUI::Session::XML::ExternalPropertyIdentifierAtt)
                                           .toString();
            variant = identifier;
        }
    } else if (parameter_type == "ComboProperty") {
        QString selections =
            reader->attributes().value(GUI::Session::XML::ParameterValueAttribute).toString();
        QString values =
            reader->attributes().value(GUI::Session::XML::ParameterExtAttribute).toString();

        ComboProperty combo_property;
        combo_property.setStringOfValues(values);
        combo_property.setStringOfSelections(selections);

        variant = combo_property.variant();
    } else {
        QString message = QString("SessionModel::readProperty: parameter type not supported '"
                                  + parameter_type + "'");
        throw Error(message);
    }

    if (variant.isValid())
        item->setRoleProperty(role, variant);

    return parameter_name;
}

bool GUI::Session::XML::readBoolAttribute(QXmlStreamReader* reader, const QString& attributeName)
{
    return reader->attributes().value(attributeName).toUInt() != 0;
}

unsigned GUI::Session::XML::readUIntAttribute(QXmlStreamReader* reader,
                                              const QString& attributeName)
{
    return reader->attributes().value(attributeName).toUInt();
}

void GUI::Session::XML::readAttribute(QXmlStreamReader* reader, const QString& attributeName,
                                      QString* s)
{
    *s = reader->attributes().value(attributeName).toString();
}

void GUI::Session::XML::readAttribute(QXmlStreamReader* reader, const QString& attributeName,
                                      QColor* c)
{
    QString parameter_value = reader->attributes().value(attributeName).toString();
    *c = QColor(parameter_value);
}

void GUI::Session::XML::readAttribute(QXmlStreamReader* reader, const QString& attributeBaseName,
                                      R3* vec)
{
    double x, y, z;
    readAttribute(reader, attributeBaseName + "X", &x);
    readAttribute(reader, attributeBaseName + "Y", &y);
    readAttribute(reader, attributeBaseName + "Z", &z);
    *vec = {x, y, z};
}

void GUI::Session::XML::readAttribute(QXmlStreamReader* reader, const QString& attributeBaseName,
                                      complex_t* c)
{
    double r, i;
    readAttribute(reader, attributeBaseName + "Re", &r);
    readAttribute(reader, attributeBaseName + "Im", &i);
    c->real(r);
    c->imag(i);
}

void GUI::Session::XML::readAttribute(QXmlStreamReader* reader, const QString& attributeName,
                                      double* d)
{
    *d = reader->attributes().value(attributeName).toDouble();
}
